#define _FILE_OFFSET_BITS 64
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>



struct descriptor {
    const char *file;
    int fd;
    off_t size;
};

static struct descriptor *files;
static int nfiles;

bool do_sleep() {
    struct timespec time_req, time_rem;
    int result;

    time_req.tv_sec = 0;
    time_req.tv_nsec = 500*1000*1000;

    while (true) {
        result = nanosleep(&time_req, &time_rem);
        if (result == 0)
            return true;
        if (errno != EINTR)
            return false;

        time_req = time_rem;
    }
}

off_t get_size(struct descriptor* p) {
    struct stat statbuf;
    int result = fstat(p->fd, &statbuf);

    if (result == -1) {
        char *errstr = strerror(errno);
        fprintf(stderr, "Could not stat file '%s'\n  %s\n", p->file, errstr);
        return (off_t)-1;
    }

    return statbuf.st_size;
}

bool init_files(int argc, char *argv[]) {
    int i;
    char *errstr;

    nfiles = argc - 1;
    files = malloc(sizeof(files[0]) * nfiles);
    if (!files) {
        errstr = strerror(errno);
        fprintf(stderr, "Could not allocate memory\n  %s\n", errstr);
        return false;
    }

    for (i=0; i < nfiles; i++) {
        struct descriptor *p = &(files[i]);

        p->file = argv[i+1];
        p->fd = open(p->file, O_RDONLY | O_LARGEFILE);
        if (p->fd == -1) {
            errstr = strerror(errno);
            fprintf(stderr, "Could not open file '%s'\n  %s\n", p->file, errstr);
            return false;
        }

        p->size = get_size(p);
        if (p->size == (off_t)-1)
            return false;
    }

    return true;
}

bool do_write(int fd, const char *buf, size_t count) {
    int i;
    char *errstr;
    static int queued_NL = 0;

    while (queued_NL) {
        if (write(1, "\n", 1) == -1)
            goto error_out;
        (queued_NL)--;
    }

    for (i=count-1; buf[i] == '\n'; i--) {
        queued_NL++;
        count--;
    }

    if (write(1, buf, count) == -1)
        goto error_out;

    return true;

error_out:
    errstr = strerror(errno);
    fprintf(stderr, "Could not write to fd %d\n  %s\n", fd, errstr);
    return false;
}


bool do_cat(struct descriptor *p, off_t new_size) {
    char *errstr;
    char buf[BUFSIZ];
    ssize_t nread;

    if (lseek(p->fd, p->size, SEEK_SET) == (off_t)-1) {
        errstr = strerror(errno);
        fprintf(stderr, "Could not seek in file '%s'\n  %s\n", p->file, errstr);
        return false;
    }

    nread = read(p->fd, buf, sizeof(buf));
    while (nread && nread != -1) {
        if (!do_write(1, buf, nread))
            return false;

        nread = read(p->fd, buf, sizeof(buf));
    }

    if (nread == -1) {
        errstr = strerror(errno);
        fprintf(stderr, "Could not read from file '%s'\n  %s\n", p->file, errstr);
        return false;
    }

    return true;
}

bool print_filename(const char * const file) {
    char buf[12+strlen(file)+9+1];
    char *ptr = buf;

    ptr = stpcpy(ptr, "\n\033[1;37m==> ");
    ptr = stpcpy(ptr, file);
    ptr = stpcpy(ptr, " <==\033[0m\n");

    return do_write(1, buf, strlen(buf));
}

bool cat_files() {
    int i;
    off_t new_size;
    struct descriptor *p;
    static struct descriptor *last_printed = NULL;

    for (i=0; i < nfiles; i++) {
        p = &(files[i]);

        new_size = get_size(p);
        if (new_size == (off_t)-1)
            return false;

        if (new_size > p->size) {
            if (last_printed != p) {
                last_printed = p;
                if (!print_filename(p->file))
                    return false;
            }

            if (!do_cat(p, new_size))
                return false;
        }

        p->size = new_size;
    }

    return true;
}

int main(int argc, char* argv[]) {
    if (!init_files(argc, argv))
        return EXIT_FAILURE;
    

    while (true) {
        if (!do_sleep() || !cat_files())
            return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
